/*
 * minidns.c - miniDNS update client for unix-like systems.
 *
 * (c) 2002 Kyle Donaldson <gile@livejournal.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netdb.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <signal.h>
#include <syslog.h>
#include "md5.h"

#define MD_SERVER_PORT		9120
#define CONFIG_FILE		"/etc/minidns.conf"
#define MD_PRIO			(LOG_USER | LOG_WARNING)
/* 0.8.1 as a hex number */
#define MD_HIGH_VER		0x00000801

static int md_current_ipaddr = 0;

static struct __md_config {
	char * domains[10];
	char * user;
	char * pass;
	char * interface;
	char * server;
	int interval;
	int dct;
} config[1];

int openhost (char * host, int prt)
{
	int sock;
	struct sockaddr_in addr;
	struct hostent * he;
	he = gethostbyname(host);
	if (he == NULL)
		goto _oh_error;
	sock = socket(AF_INET, SOCK_STREAM, getprotobyname("tcp")->p_proto);
	if (sock == -1)
		goto _oh_error;
	memcpy(&addr.sin_addr, he->h_addr, he->h_length);
	addr.sin_family = AF_INET;
	addr.sin_port = htons(prt);
	if (connect(sock, (struct sockaddr *)&addr, sizeof(addr)) == -1)
		goto _oh_error;
	return sock;
_oh_error:
	syslog(MD_PRIO, "Connect to %s failed: %m.", host);
	return -1;
}

char * sock_read (int sock)
{
	char * ret;
	char * s;
	char ch;
	char id[20];
	int i;
	ret = malloc(12288);
	if (!ret)
		return NULL;
	s = ret;
	while (read(sock, &ch, 1) == 1) {
		if (ch == '\n')
			break;
		else if (ch != '\r')
			*s++ = ch;
	}
	*s = '\0';
	if (ch != '\n') {
		free(ret);
		return NULL;
	}
	for (i = 0; i < strlen(ret); i++)
		if (ret[i] == ' ')
			break;
	i++;
	strncpy(id, ret, i);
	if (!strcmp(id, "ERR")) {
		syslog(MD_PRIO, "Update error: %s", ret + i);
		free(ret);
		return NULL;
	}
	return ret + i;
}

static unsigned char md_char_int (char ch)
{
	if ((ch >= 0x30) && (ch <= 0x39))
		return ch - 0x30;
	switch (ch) {
		case 'a': case 'A': return 0x0A;
		case 'b': case 'B': return 0x0B;
		case 'c': case 'C': return 0x0C;
		case 'd': case 'D': return 0x0D;
		case 'e': case 'E': return 0x0E;
		case 'f': case 'F': return 0x0F;
	}
	return 0;
}

unsigned char * md_bin_pack (char * str)
{
	unsigned char * ret;
	int i;
	int p = 0;
	ret = malloc(17);
	if (!ret)
		return NULL;
	for (i = 0; i < 32; i += 2) {
		ret[p] = (md_char_int(str[i]) << 4);
		ret[p] += md_char_int(str[i + 1]);
		p++;
	}
	return ret;
}

char * md_dot_quad (u_int32_t addr)
{
	char * ret;
	int part[4];
	ret = malloc(16);
	if (!ret)
		return "0.0.0.0";
	part[0] = (addr >> 24);
	part[1] = ((addr >> 16) - ((addr >> 24) << 8));
	part[2] = ((addr >> 8) - ((addr >> 16) << 8));
	part[3] = ((addr) - ((addr >> 8) << 8));
	sprintf(ret, "%d.%d.%d.%d", part[0], part[1], part[2], part[3]);
	return ret;
}

char * md_command_send (int fd, char * cmd, char * data0, char * data1, char * data2)
{
	char * temp;
	char * resp;
	temp = malloc(strlen(cmd) + (data0 ? strlen(data0) : 0) + (data1 ? strlen(data1) : 0) + (data2 ? strlen(data2) : 0) + 6);
	if (!temp)
		return NULL;
	strcpy(temp, cmd);
	if (data0) {
		strcat(temp, " ");
		strcat(temp, data0);
	}
	if (data1) {
		strcat(temp, " ");
		strcat(temp, data1);
	}
	if (data2) {
		strcat(temp, " ");
		strcat(temp, data2);
	}
	strcat(temp, "\n");
	write(fd, temp, strlen(temp));
	resp = sock_read(fd);
	if (!resp) {
		syslog(MD_PRIO, "Failed command was: %s", temp);
		free(temp);
		return NULL;
	}
	free(temp);
	return resp;
}

int md_update_addr (u_int32_t ipaddr)
{
	int sock;
	int i;
	int ver = 0;
	md5_byte_t digest[16];
	md5_state_t status;
	char * chall;
	sock = openhost(config->server, MD_SERVER_PORT);
	if (sock == -1)
		return -1;
	/* welcome message */
	sock_read(sock);
	md_command_send(sock, "agent", "Unix-MiniDNS/1.1", NULL, NULL);
	chall = md_command_send(sock, "version", NULL, NULL, NULL);
	ver += strtol(chall, &chall, 10) << 16;
	chall++;
	ver += strtol(chall, &chall, 10) << 8;
	chall++;
	ver += strtol(chall, &chall, 10);
	if (ver > MD_HIGH_VER) {
		syslog(MD_PRIO, "server version is greater than supported.  consider upgrading");
	}
	chall = md_command_send(sock, "login", config->user, "digest-md5", NULL);
	md5_init(&status);
	md5_append(&status, config->pass, strlen(config->pass));
	md5_finish(&status, digest);
	md5_init(&status);
	md5_append(&status, digest, 16);
	md5_append(&status, md_bin_pack(chall), strlen(chall) / 2);
	md5_finish(&status, digest);
	free(chall);
	md_command_send(sock, "response", md5_mkstring(digest), NULL, NULL);
	for (i = 0; i < config->dct; i++) {
		md_command_send(sock, "a_update", "online", config->domains[i], md_dot_quad(ipaddr));
		syslog(MD_PRIO, "%s is now online.", config->domains[i]);
	}
	md_command_send(sock, "exit", NULL, NULL, NULL);
	close(sock);
	syslog(MD_PRIO, "IP address changed to %s.", md_dot_quad(ipaddr));
	return 0;
}

u_int32_t md_ipv4_addr (void)
{
	int sock;
	struct ifreq ifr;
	struct sockaddr_in * sdr;
	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock == -1)
		return 0;
	strcpy(ifr.ifr_name, config->interface);
	ifr.ifr_addr.sa_family = AF_INET;
	if (ioctl(sock, SIOCGIFADDR, &ifr) != 0)
		return 0;
	sdr = (struct sockaddr_in *) &ifr.ifr_addr;
	close(sock);
	return ntohl(sdr->sin_addr.s_addr);
}

int md_load_config (void)
{
	FILE * fd;
	char line[512];
	char v_name[100];
	char v_val[200];
	register int i;
	int lnc = 0;
	int dnc = 0;
	fd = fopen(CONFIG_FILE, "r");
	if (!fd) {
		fprintf(stderr, "Cannot open config file %s: %s.\n", CONFIG_FILE, strerror(errno));
		fprintf(stderr, "miniDNS cannot run without a configuration file, exiting...\n");
		return -1;
	}
	while (!feof(fd)) {
		lnc++;
		memset(line, 0, 512);
		memset(v_name, 0, 100);
		memset(v_val, 0, 200);
		fgets(line, 512, fd);
		if (strlen(line) < 5)
			continue;
		line[strlen(line) - 1] = '\0';
		for (i = 0; (isspace(line[i])) && (i < strlen(line)); i++);
		if (i > 0)
			strcpy(line, line + i);
		if (line[0] == '#')
			continue;
		for (i = 0; (!isspace(line[i])) && (i < strlen(line)); i++);
		strncpy(v_name, line, i);
		strcpy(line, line + i);
		for (i = 0; (isspace(line[i])) && (i < strlen(line)); i++);
		if (line[i] != '=') {
			fprintf(stderr, "%s:%d: Missing '=' separator.  Aborting...\n", CONFIG_FILE, lnc);
			fclose(fd);
			return -1;
		}
		for (i++; (isspace(line[i])) && (i < strlen(line)); i++);
		strcpy(v_val, line + i);
		if (!strcasecmp(v_name, "Username"))
			config->user = strdup(v_val);
		else if (!strcasecmp(v_name, "Password"))
			config->pass = strdup(v_val);
		else if (!strcasecmp(v_name, "Update-Interval"))
			config->interval = atoi(v_val) * 60;
		else if (!strcasecmp(v_name, "Update-Server"))
			config->server = strdup(v_val);
		else if (!strcasecmp(v_name, "Network-Interface"))
			config->interface = strdup(v_val);
		else if (!strcasecmp(v_name, "Domain-Name")) {
			if (dnc >= 10) {
				fprintf(stderr, "Ignoring domain %s, more than 10 already specified\n", v_val);
				continue;
			}
			config->domains[dnc] = strdup(v_val);
			dnc++;
		}
	}
	config->dct = dnc;
	fclose(fd);
	return 0;
}

void md_signal (int sig)
{
	switch (sig) {
		case SIGHUP:
			syslog(MD_PRIO, "reloading config file (active on next update)");
			md_load_config();
			break;
		default: break;
	}
}

int main (int argc, char ** argv)
{
	u_int32_t ipaddr;
	signal(SIGHUP, md_signal);
	openlog("miniDNS", LOG_CONS | LOG_NDELAY | LOG_PID, LOG_USER);
	config->server = "update.minidns.net";
	config->interval = 3600;
	if (md_load_config() != 0)
		exit(1);
	daemon(0, 0);
	syslog(MD_PRIO, "miniDNS/1.1 now running with %d domains.", config->dct);
	while (1) {
		ipaddr = md_ipv4_addr();
		if (md_current_ipaddr != ipaddr) {
			if (md_update_addr(ipaddr) != 0)
				/*exit(1)*/;
			md_current_ipaddr = ipaddr;
		}
		sleep(config->interval);
	}
	closelog();
	exit(1);
}
